import tempfile
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, call, patch, seal

from django.test import override_settings

import docker
from arkindex.ponos.models import Farm, State, Task
from arkindex.ponos.tasks import run_docker_task
from arkindex.process.models import ProcessMode
from arkindex.project.tests import FixtureTestCase


class TestRunDockerTask(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.first()
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            corpus=cls.corpus,
            farm=cls.farm,
        )
        cls.task = cls.process.tasks.create(
            slug="something",
            image="image",
            run=0,
            depth=0,
            state=State.Pending,
        )

    def test_local_image(self):
        client_mock = MagicMock()
        # Pulling the image fails
        client_mock.images.pull.side_effect = docker.errors.ImageNotFound("Oh no")
        # An image is returned when listing local images
        client_mock.images.list.return_value = [docker.models.images.Image()]
        # End in a container conflict just to stop the function early, because we don't need to test beyond that
        client_mock.containers.get.return_value = None
        seal(client_mock)

        with (
            tempfile.TemporaryDirectory() as temp_dir,
            self.assertRaisesMessage(ValueError, f"A Docker container named ponos-{self.task.id} already exists"),
        ):
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.images.list.call_count, 1)
        self.assertEqual(client_mock.images.list.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))

    def test_image_not_found(self):
        client_mock = MagicMock()
        # Pulling the image fails
        client_mock.images.pull.side_effect = docker.errors.ImageNotFound("Oh no")
        # Listing images locally finds nothing
        client_mock.images.list.return_value = []
        seal(client_mock)

        with (
            tempfile.TemporaryDirectory() as temp_dir,
            self.assertRaisesMessage(Exception, "Image not found locally nor remotely: Oh no"),
        ):
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.images.list.call_count, 1)
        self.assertEqual(client_mock.images.list.call_args, call("image"))

    def test_conflict(self):
        client_mock = MagicMock()
        client_mock.images.pull.return_value = None
        client_mock.containers.get.return_value = None
        seal(client_mock)

        with (
            tempfile.TemporaryDirectory() as temp_dir,
            self.assertRaisesMessage(ValueError, f"A Docker container named ponos-{self.task.id} already exists"),
        ):
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))

    def test_artifact_out_of_bounds(self):
        parent_task = self.process.tasks.create(
            slug="parent",
            image="image",
            depth=0,
            run=0,
            state=State.Completed,
        )
        self.task.depth = 1
        self.task.save()
        self.task.parents.add(parent_task)
        parent_task.artifacts.create(
            content_type="text/plain",
            path="../../../../../etc/shadow",
            size=1e9,
        )

        client_mock = MagicMock()
        client_mock.images.pull.return_value = None
        seal(client_mock)

        with (
            tempfile.TemporaryDirectory() as temp_dir,
            self.assertRaisesMessage(AssertionError, "Invalid artifact path: ../../../../../etc/shadow.")
        ):
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))

    @patch("arkindex.ponos.tasks.download_extra_files")
    def test_extra_files_error(self, download_extra_files_mock):
        self.task.extra_files = {"something": "https://teklia.com/dev/null"}
        self.task.save()

        client_mock = MagicMock()
        client_mock.images.pull.return_value = None
        seal(client_mock)

        download_extra_files_mock.side_effect = ZeroDivisionError
        seal(download_extra_files_mock)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(download_extra_files_mock.call_count, 1)
        self.assertEqual(download_extra_files_mock.call_args, call(self.task, Path(temp_dir)))

        self.task.refresh_from_db()
        self.assertEqual(self.task.state, State.Error)

    @override_settings(PONOS_DOCKER_AUTO_REMOVE_CONTAINER=False)
    def test_gpu(self):
        self.task.requires_gpu = True
        self.task.save()

        client_mock = MagicMock()
        client_mock.images.pull.return_value = None
        client_mock.containers.get.side_effect = docker.errors.NotFound("Not found.")
        # Make the function stop when the container runs, as we don't need to test any further
        client_mock.containers.run.side_effect = KeyboardInterrupt
        seal(client_mock)

        with (
            tempfile.TemporaryDirectory() as temp_dir,
            self.assertRaises(KeyboardInterrupt)
        ):
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))
        self.assertEqual(client_mock.containers.run.call_count, 1)
        self.assertEqual(client_mock.containers.run.call_args, call(
            "image",
            environment={"PONOS_DATA": "/data"},
            detach=True,
            network="host",
            volumes={temp_dir: {"bind": "/data", "mode": "rw"}},
            name=f"ponos-{self.task.id}",
            device_requests=[
                docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
            ]
        ))

    @override_settings(PONOS_DOCKER_AUTO_REMOVE_CONTAINER=True)
    @patch("arkindex.ponos.models.Artifact.s3_object")
    @patch("arkindex.ponos.tasks.upload_logs")
    @patch("arkindex.ponos.tasks.sleep")
    def test_completed(self, sleep_mock, upload_logs_mock, artifact_s3_object_mock):
        self.task.command = "format c:"
        self.task.env = {"PATH": "/dev/null"}
        self.task.save()

        client_mock = MagicMock()
        container = client_mock.containers.run.return_value

        # The first call occurs while the task is running, the second after it has finished.
        container.logs.side_effect = [b"Running", b"Done"]

        # This will be accessed after the container has finished
        container.attrs = {"State": {"ExitCode": 0}}

        def add_artifacts(*args):
            # sleep() is called after one iteration of the running task loop is complete.
            # We use this mock to not wait during tests, check that the task is properly running,
            # and create artifacts that the function should upload once the container exists.
            self.task.refresh_from_db()
            self.assertEqual(self.task.state, State.Running)
            (Path(temp_dir) / str(self.task.id) / "something.txt").write_text("blah")

        # Set up all the remaining attributes, then seal the mocks so nonexistent attributes can't be accessed
        sleep_mock.side_effect = add_artifacts
        seal(sleep_mock)

        client_mock.images.pull.return_value = None
        client_mock.containers.get.side_effect = docker.errors.NotFound("Not found.")
        container.reload.return_value = None
        container.remove.return_value = None
        # Sealing is not friendly with PropertyMocks, so we put a placeholder first
        container.status = None
        seal(client_mock)
        # Limit the amount of times container.status can be accessed, so we can't get stuck in an infinite loop
        type(container).status = PropertyMock(side_effect=[
            "running", # Loop that checks whether the container is `created` and needs to be awaited
            "running", # First iteration of the running task loop
            "exited", # Second iteration where the task is completed
        ])

        upload_logs_mock.return_value = None
        seal(upload_logs_mock)

        artifact_s3_object_mock.upload_file.return_value = None
        seal(artifact_s3_object_mock)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.task.refresh_from_db()
        self.assertEqual(self.task.state, State.Completed)

        artifact = self.task.artifacts.get()
        self.assertEqual(artifact.path, "something.txt")
        self.assertEqual(artifact.content_type, "text/plain")
        self.assertEqual(artifact.size, 4)

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))
        self.assertEqual(client_mock.containers.run.call_count, 1)
        self.assertEqual(client_mock.containers.run.call_args, call(
            "image",
            command="format c:",
            environment={"PATH": "/dev/null", "PONOS_DATA": "/data"},
            detach=True,
            network="host",
            volumes={temp_dir: {"bind": "/data", "mode": "rw"}},
            name=f"ponos-{self.task.id}",
        ))

        self.assertEqual(container.reload.call_count, 2)
        self.assertEqual(container.reload.call_args_list, [call(), call()])
        self.assertEqual(container.logs.call_count, 2)
        self.assertEqual(container.logs.call_args_list, [call(), call()])
        self.assertEqual(container.remove.call_count, 1)
        self.assertEqual(container.remove.call_args, call())

        self.assertEqual(upload_logs_mock.call_count, 2)
        self.assertEqual(upload_logs_mock.call_args_list, [call(self.task, b"Running"), call(self.task, b"Done")])

        self.assertEqual(sleep_mock.call_count, 1)
        self.assertEqual(sleep_mock.call_args, call(1))

        self.assertEqual(artifact_s3_object_mock.upload_file.call_count, 1)
        self.assertEqual(artifact_s3_object_mock.upload_file.call_args, call(
            str(Path(temp_dir) / str(self.task.id) / "something.txt"),
            ExtraArgs={"ContentType": "text/plain"},
        ))

    @override_settings(PONOS_DOCKER_AUTO_REMOVE_CONTAINER=False)
    @patch("arkindex.ponos.utils.upload_artifact")
    @patch("arkindex.ponos.tasks.upload_logs")
    @patch("arkindex.ponos.tasks.sleep")
    def test_failed(self, sleep_mock, upload_logs_mock, upload_artifact_mock):
        client_mock = MagicMock()
        container = client_mock.containers.run.return_value

        # The first call occurs while the task is running, the second after it has finished.
        container.logs.side_effect = [b"Running", b"(whilhelm scream)"]

        # This will be accessed after the container has failed
        container.attrs = {"State": {"ExitCode": 1}}

        def add_artifacts(*args):
            # sleep() is called after one iteration of the running task loop is complete.
            # We use this mock to not wait during tests, check that the task is properly running,
            # and create artifacts that the function should upload once the container exists.
            self.task.refresh_from_db()
            self.assertEqual(self.task.state, State.Running)
            # This artifact should never be uploaded
            (Path(temp_dir) / str(self.task.id) / "something.txt").write_text("blah")

        # Set up all the remaining attributes, then seal the mocks so nonexistent attributes can't be accessed
        sleep_mock.side_effect = add_artifacts
        seal(sleep_mock)

        client_mock.images.pull.return_value = None
        client_mock.containers.get.side_effect = docker.errors.NotFound("Not found.")
        container.reload.return_value = None
        # Sealing is not friendly with PropertyMocks, so we put a placeholder first
        container.status = None
        seal(client_mock)
        # Limit the amount of times container.status can be accessed, so we can't get stuck in an infinite loop
        type(container).status = PropertyMock(side_effect=[
            "running", # Loop that checks whether the container is `created` and needs to be awaited
            "running", # First iteration of the running task loop
            "exited", # Second iteration where the task has failed
        ])

        upload_logs_mock.return_value = None
        seal(upload_logs_mock)

        # We only mock this so that we can make sure we never upload any artifact
        seal(upload_artifact_mock)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.task.refresh_from_db()
        self.assertEqual(self.task.state, State.Failed)

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))
        self.assertEqual(client_mock.containers.run.call_count, 1)
        self.assertEqual(client_mock.containers.run.call_args, call(
            "image",
            environment={"PONOS_DATA": "/data"},
            detach=True,
            network="host",
            volumes={temp_dir: {"bind": "/data", "mode": "rw"}},
            name=f"ponos-{self.task.id}",
        ))

        self.assertEqual(container.reload.call_count, 2)
        self.assertEqual(container.reload.call_args_list, [call(), call()])
        self.assertEqual(container.logs.call_count, 2)
        self.assertEqual(container.logs.call_args_list, [call(), call()])

        self.assertEqual(upload_logs_mock.call_count, 2)
        self.assertEqual(upload_logs_mock.call_args_list, [call(self.task, b"Running"), call(self.task, b"(whilhelm scream)")])

        self.assertEqual(sleep_mock.call_count, 1)
        self.assertEqual(sleep_mock.call_args, call(1))

        self.assertEqual(upload_artifact_mock.call_count, 0)


    @override_settings(PONOS_DOCKER_AUTO_REMOVE_CONTAINER=True)
    @patch("arkindex.ponos.utils.upload_artifact")
    @patch("arkindex.ponos.tasks.upload_logs")
    @patch("arkindex.ponos.tasks.sleep")
    def test_stop(self, sleep_mock, upload_logs_mock, upload_artifact_mock):
        self.task.env = {"PATH": "/dev/null"}
        self.task.save()

        client_mock = MagicMock()
        container = client_mock.containers.run.return_value

        # The first two calls occur while the task is running and should not result in any upload.
        # Second call occurs after the container has been stopped.
        container.logs.side_effect = [b"", b"", b"^C"]

        # This will be accessed after the container has been stopped
        container.attrs = {
            "State": {
                # Stopping a container will likely cause a non-zero code, but it should be ignored
                # and the task should be marked as Stopped anyway
                "ExitCode": 1,
            }
        }

        def set_stopping(*args):
            # sleep() is called after one iteration of the running task loop is complete.
            # We use this mock to both not wait during tests and mark the task as stopping.
            self.task.refresh_from_db()
            self.assertEqual(self.task.state, State.Running)
            # We use .update() and not Task.save to ensure run_docker_task properly reloads the task from the DB.
            Task.objects.filter(id=self.task.id).update(state=State.Stopping)

        # Set up all the remaining attributes, then seal the mocks so nonexistent attributes can't be accessed
        sleep_mock.side_effect = set_stopping
        seal(sleep_mock)

        client_mock.images.pull.return_value = None
        client_mock.containers.get.side_effect = docker.errors.NotFound("Not found.")
        container.reload.return_value = None
        container.stop.return_value = None
        container.remove.return_value = None
        # Sealing is not friendly with PropertyMocks, so we put a placeholder first
        container.status = None
        seal(client_mock)
        # Limit the amount of times container.status can be accessed, so we can't get stuck in an infinite loop
        type(container).status = PropertyMock(side_effect=[
            "running", # Loop that checks whether the container is `created` and needs to be awaited
            "running", # First iteration of the running task loop
            "running", # Second iteration where the task is Stopping
        ])

        upload_logs_mock.return_value = None
        seal(upload_logs_mock)

        # We only mock this so that we can make sure we never upload any artifact
        seal(upload_artifact_mock)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_docker_task(client_mock, self.task, Path(temp_dir))

        self.task.refresh_from_db()
        self.assertEqual(self.task.state, State.Stopped)

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.get.call_count, 1)
        self.assertEqual(client_mock.containers.get.call_args, call(f"ponos-{self.task.id}"))
        self.assertEqual(client_mock.containers.run.call_count, 1)
        self.assertEqual(client_mock.containers.run.call_args, call(
            "image",
            environment={"PATH": "/dev/null", "PONOS_DATA": "/data"},
            detach=True,
            network="host",
            volumes={temp_dir: {"bind": "/data", "mode": "rw"}},
            name=f"ponos-{self.task.id}",
        ))

        self.assertEqual(container.reload.call_count, 2)
        self.assertEqual(container.reload.call_args_list, [call(), call()])
        self.assertEqual(container.stop.call_count, 1)
        self.assertEqual(container.stop.call_args, call())
        self.assertEqual(container.logs.call_count, 3)
        self.assertEqual(container.logs.call_args_list, [call(), call(), call()])
        self.assertEqual(container.remove.call_count, 1)
        self.assertEqual(container.remove.call_args, call())

        self.assertEqual(upload_logs_mock.call_count, 1)
        self.assertEqual(upload_logs_mock.call_args, call(self.task, b"^C"))

        self.assertEqual(sleep_mock.call_count, 1)
        self.assertEqual(sleep_mock.call_args, call(1))

        self.assertEqual(upload_artifact_mock.call_count, 0)
