import uuid
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.db import transaction
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import (
    FeatureUsage,
    ProcessMode,
    WorkerRun,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role


class TestWorkerRunsCreate(FixtureAPITestCase):
    """
    Test worker runs create endpoint
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.local_process = cls.user.processes.get(mode=ProcessMode.Local)
        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1, parents=[])
        cls.configuration_1 = cls.worker_1.configurations.create(name="My config", configuration={"key": "value"})
        worker_version = WorkerVersion.objects.exclude(worker=cls.version_1.worker).first()
        cls.configuration_2 = worker_version.worker.configurations.create(name="Config")
        cls.process_2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)

        # Model and Model version setup
        cls.model_1 = Model.objects.create(name="My model")
        cls.model_1.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.model_version_1 = ModelVersion.objects.create(
            model=cls.model_1,
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )
        cls.version_gpu_required = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "ikari shinji"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Required,
            version=2
        )
        cls.version_gpu_supported = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "ayanami rei"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Supported,
            version=3
        )
        cls.version_gpu_disabled = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "soryu asuka langley"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Disabled,
            version=4
        )

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_create_no_execution_right(self, max_level_mock):
        """
        An execution access on the target worker version is required to create a worker run
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["You do not have an execution access to this worker."]})

        self.assertListEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1)
        ])

    def test_create_invalid_version(self):
        self.client.force_login(self.user)
        version_id = uuid.uuid4()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": version_id},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker_version_id": [f'Invalid pk "{version_id}" - object does not exist.'],
        })

    def test_create_unique(self):
        self.client.force_login(self.user)
        self.version_1.model_usage = FeatureUsage.Required
        self.version_1.save()
        cases = [
            (None, None),
            (None, self.configuration_1),
            (self.model_version_1, None),
            (self.model_version_1, self.configuration_1),
        ]
        for model_version, configuration in cases:
            with self.subTest(model_version=model_version, configuration=configuration):
                self.run_1.model_version = model_version
                self.run_1.configuration = configuration
                self.run_1.save()

                # Having a model version or a configuration adds one query for each
                query_count = 5 + bool(model_version) + bool(configuration)

                with self.assertNumQueries(query_count):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}),
                        data={
                            "worker_version_id": str(self.version_1.id),
                            "model_version_id": str(model_version.id) if model_version else None,
                            "configuration_id": str(configuration.id) if configuration else None,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "__all__": ["A WorkerRun already exists on this process with the selected worker version, model version and configuration."],
                })

    def test_create_unavailable_version(self):
        self.version_1.state = WorkerVersionState.Error
        self.version_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["This WorkerVersion is not in an Available state."]})

    def test_create_archived_worker(self):
        self.worker_1.archived = datetime.now(timezone.utc)
        self.worker_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["This WorkerVersion is part of an archived worker."]})

    def test_create_archived_model(self):
        self.model_1.archived = datetime.now(timezone.utc)
        self.model_1.save()
        self.version_1.model_usage = FeatureUsage.Supported
        self.version_1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "model_version_id": str(self.model_version_1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"model_version_id": ["This ModelVersion is part of an archived model."]})

    def test_create_invalid_process_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": uuid.uuid4()}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Process matches the given query."})

    def test_create_invalid_process_mode(self):
        self.client.force_login(self.user)

        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process_2.mode = mode
                self.process_2.save()

                with self.assertNumQueries(5):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                        {"worker_version_id": str(self.version_1.id)},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."],
                })

    def test_create_non_corpus_process_mode(self):
        process = self.user.processes.get(mode=ProcessMode.Local)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(process.id)}),
                {"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_process_already_started(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            farm=self.farm,
        )
        process.run()

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(process.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": []},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns cannot be added or updated on processes that have already started."],
        })

    def test_create(self):
        self.client.force_login(self.user)

        for mode in (ProcessMode.Workers, ProcessMode.Dataset):
            self.process_2.worker_runs.filter(version=self.version_1).delete()

            with self.subTest(mode=mode):
                self.process_2.mode = mode
                self.process_2.save()

                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                        {"worker_version_id": str(self.version_1.id), "parents": []},
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                pk = data.pop("id")
                self.assertNotEqual(pk, self.run_1.id)
                self.assertDictEqual(data, {
                    "worker_version": {
                        "id": str(self.version_1.id),
                        "configuration": {"test": 42},
                        "docker_image_iid": self.version_1.docker_image_iid,
                        "gpu_usage": "disabled",
                        "model_usage": FeatureUsage.Disabled.value,
                        "revision_url": None,
                        "version": 1,
                        "tag": None,
                        "branch": None,
                        "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                        "state": "available",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer",
                            "description": "",
                            "repository_url": None,
                            "archived": False,
                        }
                    },
                    "parents": [],
                    "model_version": None,
                    "configuration": None,
                    "process": {
                        "id": str(self.process_2.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": mode.value,
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                    },
                    "use_gpu": False,
                    "summary": "Worker Recognizer @ version 1",
                })
                run = WorkerRun.objects.get(pk=pk)
                # Check generated summary
                self.assertEqual(run.summary, "Worker Recognizer @ version 1")

    def test_create_empty(self):
        """
        The worker_version_id is required to create a worker run
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_version_id": ["This field is required."],
        })

    def test_create_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={
                    "worker_version_id": str(self.version_1.id),
                    "parents": [],
                    "configuration_id": str(self.configuration_1.id)
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop("id")
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "process": {
                "id": str(self.process_2.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1 using configuration 'My config'",
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary
        self.assertEqual(run.summary, "Worker Recognizer @ version 1 using configuration 'My config'")

    def test_create_invalid_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                data={"worker_version_id": str(self.version_1.id), "parents": [], "configuration_id": str(self.configuration_2.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration_id": ["The configuration must be part of the same worker."]})

    def test_create_summary(self):
        self.client.force_login(self.user)
        test_version = self.worker_1.versions.create(
            version=5,
            state=WorkerVersionState.Available,
            docker_image_iid="registry.gitlab.com/dead-sea-scrolls:004",
            model_usage=FeatureUsage.Supported
        )

        cases = [
            ("eva-01", None, "revision_url", self.model_version_1, self.configuration_1, f"Worker Recognizer @ eva-01 ({str(test_version.id)[:6]}) with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'"),
            (None, None, "revision_url", self.model_version_1, self.configuration_1, f"Worker Recognizer @ {str(test_version.id)[:6]} with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'"),
            ("eva-01", None, "revision_url", None, self.configuration_1, f"Worker Recognizer @ eva-01 ({str(test_version.id)[:6]}) using configuration 'My config'"),
            (None, None, "revision_url", self.model_version_1, None, f"Worker Recognizer @ {str(test_version.id)[:6]} with model My model @ {str(self.model_version_1.id)[:6]}"),
            ("eva-01", None, "revision_url", None, None, f"Worker Recognizer @ eva-01 ({str(test_version.id)[:6]})"),
            (None, None, "revision_url", None, None, f"Worker Recognizer @ {str(test_version.id)[:6]}"),
            ("eva-01", None, None, self.model_version_1, self.configuration_1, f"Worker Recognizer @ eva-01 (version 5) with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'"),
            (None, None, None, self.model_version_1, self.configuration_1, f"Worker Recognizer @ version 5 with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'"),
            ("eva-01", None, None, None, self.configuration_1, "Worker Recognizer @ eva-01 (version 5) using configuration 'My config'"),
            (None, None, None, self.model_version_1, None, f"Worker Recognizer @ version 5 with model My model @ {str(self.model_version_1.id)[:6]}"),
            ("eva-01", None, None, None, None, "Worker Recognizer @ eva-01 (version 5)"),
            ("eva-01", "operation-yashima", None, None, None, "Worker Recognizer @ eva-01 (version 5)"),
            (None, "operation-yashima", None, None, None, "Worker Recognizer @ operation-yashima (version 5)"),
            (None, None, None, None, None, "Worker Recognizer @ version 5"),
        ]

        for tag, branch, revision_url, model_version, config, expected_summary in cases:
            with self.subTest(tag=tag, branch=branch, model_version=model_version, config=config), transaction.atomic():
                # Clear the process of worker runs
                self.process_2.worker_runs.all().delete()

                num_queries = 6

                test_version.version = None
                test_version.tag = tag
                test_version.branch = branch
                test_version.revision_url = revision_url
                if not revision_url:
                    test_version.version = 5
                test_version.save()

                payload = {
                    "worker_version_id": str(test_version.id),
                    "parents": [],
                }
                if model_version:
                    payload["model_version_id"] = model_version.id
                    # If there is a model version, it adds a query
                    num_queries += 1
                if config:
                    payload["configuration_id"] = config.id
                    # If there is a configuration, it adds a query
                    num_queries += 1

                with self.assertNumQueries(num_queries):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_2.id)}),
                        data=payload,
                    )
                    if response.status_code == 400:
                        self.assertDictEqual(response.json(), {})
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                created_run = WorkerRun.objects.get(pk=response.json()["id"])
                self.assertEqual(created_run.summary, expected_summary)

    def test_create_gpu_usage_auto(self):
        """
        use_gpu on the worker run is set automatically depending on the worker version's gpu_usage
        """
        self.client.force_login(self.user)

        cases = [
            (self.version_gpu_disabled, False),
            (self.version_gpu_supported, False),
            (self.version_gpu_required, True)
        ]

        for worker_version, use_gpu in cases:
            with self.subTest(worker_version=worker_version, use_gpu=use_gpu):
                with self.assertNumQueries(6):
                    response = self.client.post(
                        reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}),
                        {"worker_version_id": str(worker_version.id), "parents": []},
                    )
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                pk = data.pop("id")
                self.assertDictEqual(data, {
                    "worker_version": {
                        "id": str(worker_version.id),
                        "configuration": worker_version.configuration,
                        "docker_image_iid": "registry.nerv.co.jp/neon-genesis:evangelion",
                        "gpu_usage": worker_version.gpu_usage.value,
                        "model_usage": FeatureUsage.Disabled.value,
                        "revision_url": None,
                        "version": worker_version.version,
                        "tag": None,
                        "branch": None,
                        "created": worker_version.created.isoformat().replace("+00:00", "Z"),
                        "state": "available",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer",
                            "description": "",
                            "repository_url": None,
                            "archived": False,
                        }
                    },
                    "parents": [],
                    "model_version": None,
                    "configuration": None,
                    "process": {
                        "id": str(self.process_1.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": "workers",
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                    },
                    "summary": f"Worker Recognizer @ version {worker_version.version}",
                    "use_gpu": use_gpu
                })
                run = WorkerRun.objects.get(pk=pk)
                self.assertEqual(run.use_gpu, use_gpu)

    def test_create_use_gpu_ignored(self):
        """
        if "use_gpu" is being sent when creating a worker run, it is ignored
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:worker-run-list", kwargs={"pk": str(self.process_1.id)}),
                {"worker_version_id": str(self.version_gpu_required.id), "parents": [], "use_gpu": False},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop("id")
        self.assertDictEqual(data, {
            "worker_version": {
                "id": str(self.version_gpu_required.id),
                "configuration": {"pilot": "ikari shinji"},
                "docker_image_iid": "registry.nerv.co.jp/neon-genesis:evangelion",
                "gpu_usage": FeatureUsage.Required.value,
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 2,
                "tag": None,
                "branch": None,
                "created": self.version_gpu_required.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "parents": [],
            "model_version": None,
            "configuration": None,
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": "Worker Recognizer @ version 2",
            "use_gpu": True
        })
        run = WorkerRun.objects.get(pk=pk)
        self.assertEqual(run.use_gpu, True)
