from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import FeatureUsage, ProcessMode, Worker, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase


class TestWorkerVersionPartialUpdate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")

        cls.version_1 = cls.worker_reco.versions.get()
        cls.version_2 = cls.worker_dla.versions.get()

        farm = Farm.objects.first()
        process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=farm,
            name="test_process"
        )
        process.run()
        cls.task = process.tasks.get()

    def test_requires_login(self):
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_non_task(self):
        """
        Only Ponos tasks are allowed to update a worker version
        """
        docker_image_iid = self.version_1.docker_image_iid
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": str(docker_image_iid),
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Only Ponos tasks are allowed to update a worker version."})

    def test_partial_update(self):
        """
        Update worker version artifact, configuration, state and tag
        Updating the branch should reset the branch on any other version on the same worker
        """
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()

        self.assertNotEqual(self.version_1.worker_id, self.version_2.worker_id)
        self.version_2.branch = "operation-yashima"
        self.version_2.save()

        branch_version = self.version_1.worker.versions.create(branch="operation-yashima", version=2)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "docker_image_iid": "eva:unit-01",
                    "state": "error",
                    "tag": "eva-01",
                    "branch": "operation-yashima",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))

        self.version_1.refresh_from_db()
        self.assertEqual(self.version_1.configuration, {"test": "test2"})
        self.assertEqual(self.version_1.docker_image_iid, "eva:unit-01")
        self.assertEqual(self.version_1.state, WorkerVersionState.Error)
        self.assertEqual(self.version_1.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(self.version_1.tag, "eva-01")
        self.assertEqual(self.version_1.branch, "operation-yashima")

        self.version_2.refresh_from_db()
        self.assertEqual(self.version_2.branch, "operation-yashima")

        branch_version.refresh_from_db()
        self.assertIsNone(branch_version.branch)

    def test_unique_tag(self):
        self.worker_reco.versions.create(
            docker_image_iid="registry.fake.com/workers/worker:12345",
            state=WorkerVersionState.Available,
            version=12,
            tag="eva-01"
        )

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "tag": "eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["A version already exists for this worker with this tag."]})

    def test_empty_tag_branch(self):
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "tag": "",
                    "branch": "",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "tag": ["This field may not be blank."],
            "branch": ["This field may not be blank."],
        })

    def test_revision_ignored(self):
        self.version_1.revision_url = "url"
        self.version_1.version = None
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "revision_url": "updated"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        # revision_url just gets ignored
        self.assertEqual(self.version_1.revision_url, "url")

    def test_revision_url_ignored(self):
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            # revision_url just gets ignored
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(self.version_1.revision_url, None)

    def test_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()

        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={"state": "available"},
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })
